#!/bin/bash

## Exit function with error output - called by different errors in script
die () {
    printf >&2 "$@"
    exit 1
}

## Set some global variables for the script - working directory, script file name etc.
me=$(basename $0)
script=$(readlink -f $0)
scriptDir=$(dirname ${script})
workingDir=$(pwd)
cd ${scriptDir}

## Display current date and time on the console
date

## Check if script was run without command line arguments and display correct syntax if so
[ "$#" -eq 0 ] && die "No arguments provided!\n\nSyntax: $me [OPTIONS] /path/to/configuration/file\nAvailable [OPTIONS]:\n-d : Debug mode.\n-last-month : Download the last month of logs (use only manually as an initial download).\n-last-week : Download the last week of logs (use only manually as an initial download).\n-last-day : Download the last day of logs (use only manually as an initial download).\n"

## Check which command line arguments were set and create the corresponding runtime variables
for arg in $@
do
	[ "$arg" == "-d" ] && debugMode=1
	[ "$arg" == "-last-month" ] && startDate=$(echo $(($(date -d$(date --iso-8601='hours' --date='last month') +%s)))000)
	[ "$arg" == "-last-week" ] && startDate=$(echo $(($(date -d$(date --iso-8601='hours' --date='last week') +%s)))000)
	[ "$arg" == "-last-day" ] && startDate=$(echo $(($(date -d$(date --iso-8601='hours' --date='yesterday') +%s)))000)
done

## Define the configuration file given on the command line
confFile=$arg

## If no configuration file was given at the command line display correct syntax
[ -e "$arg" ] || die "Configuration file provided does not exist!\n\nSyntax: $me [OPTIONS] /path/to/configuration/file\nAvailable [OPTIONS]:\n-d : Debug mode.\n"

## Check availability of required tools
[ `which gawk` ] || die "Command \"gawk\" missing or not in PATH!"
[ `which sed` ] || die "Command \"sed\" missing or not in PATH!"
[ `which unzip` ] || die "Command \"unzip\" missing or not in PATH!"
[ `which find` ] || die "Command \"find\" missing or not in PATH!"
[ `which zcat` ] || die "Command \"zcat\" missing or not in PATH!"
[ `which wget` ] && w_client="wget"
[ `which curl` ] && w_client="curl"
[ ! -z "$w_client" ] || die "Command \"wget\" or \"curl\" missing or not in PATH!"
[ ! -z "$debugMode" ] && echo "### Debug mode activated"

## Disable proxy usage by default - will be re-enabled further down if set in configuration file
useProxy="0"

## Load configuration file
source $confFile
[ ! -z "$debugMode" ] && echo "### Loading configuration file: ${confFile}"

## Check if a sync token was sent in the previous request
[ -e "${destDownloadDir}/lasttoken" ] && tokenExists=$(cat ${destDownloadDir}/lasttoken)
OIFS=$IFS

## If startDate was not set as a command line argument, set it to the last full hour - syncAPI require timestamps to be set to the full hour
[ -z "$startDate" ] && startDate=$(echo $(($(date -d$(date --iso-8601='hours') +%s)-3600))000)

## Set endDate to current time in milliseconds
endDate=$(echo $(date +%s)000)

## Set variables for finding and later deleting log archives
logArchives="*.zip"
logFiles="${destDownloadDir}/*.log.gz"

## Download cloud logs
cd ${destDownloadDir}
## If API credentials were provided in configuration file, download logs from cloud. Otherwise process local log files.
if [[ ${apiUsername} != "" && ${apiPassword} != "" ]]; then
	### Download new log files from cloud
	[ ! -z "$debugMode" ] && echo "${w_client} being used to download logs"; echo "Download logs up to ${endDate}"; echo "Downloading zipped log files to ${destDownloadDir}"
	### If previous request set a sync token set it for the current download
	[ -z "$tokenExists" ] && syncToken="" || syncToken=${tokenExists}
	### Set defaults for continuous download
	syncStatus="more"
	httpStatus="200"
	### As long as the defaults do not change continue requesting log files
	while [[ "$syncStatus" == "more" && "$httpStatus" == "200" ]]
	do
		lastToken=${syncToken}
		### Download without a sync token if none was provided in previous request. Otherwise use sync token in request
		if [[ "$lastToken" == "" ]]; then
			logDownloadURL="https://portal.threatpulse.com/reportpod/logs/sync?apiUsername=${apiUsername}&apiPassword=${apiPassword}&startDate=${startDate}&endDate=0&token=none"
		else
			logDownloadURL="https://portal.threatpulse.com/reportpod/logs/sync?apiUsername=${apiUsername}&apiPassword=${apiPassword}&startDate=${startDate}&endDate=0&token=${syncToken}"
		fi
		### If wget is available, use it to download. Otherwise use curl.
		if [[ ${w_client} == "wget" ]]; then
			if [[ ${enableProxy} == "yes" ]]; then
				[ ! -z "$debugMode" ] && echo "### Using proxy: ${proxyAddress}:${proxyPort} for download"
				if [[ ${proxyAuth_User} != "" ]]; then
					wget -T 10 --no-check-certificate --server-response -o log.txt -O logs.zip -e https_proxy="http://${proxyAddress}:${proxyPort}" --proxy-user="${proxyAuth_User}" --proxy-password="${proxyAuth_Password}" "${logDownloadURL}"
				else
					wget -T 10 --no-check-certificate --server-response -o log.txt -O logs.zip -e https_proxy="http://${proxyAddress}:${proxyPort}" "${logDownloadURL}"
				fi
			else
				wget -T 10 --no-check-certificate --server-response -o log.txt -O logs.zip "${logDownloadURL}"
			fi
			sed -i -e 's/^[ \t]*//' log.txt
			grep -e 'HTTP/' -e ': ' log.txt |grep -v 'ERROR'|grep -v 'Saving to'|grep -v 'Length: unspecified' > headers.txt
			rm -f log.txt
		else
			if [[ ${enableProxy} == "yes" ]]; then
				[ ! -z "$debugMode" ] && echo "### Using proxy: ${proxyAddress}:${proxyPort} for download"
				if [[ ${proxyAuth_User} != "" ]]; then
					curl -k -D headers.txt -o logs.zip --proxy "http://${proxyAddress}:${proxyPort}" --proxy-user "${proxyAuth_User}:${proxyAuth_Password}" "${logDownloadURL}"
				else
					curl -k -D headers.txt -o logs.zip --proxy "http://${proxyAddress}:${proxyPort}" "${logDownloadURL}"
				fi
			else
				curl -k -D headers.txt -o logs.zip "${logDownloadURL}"
			fi
		fi
		### Capture syncAPI trailer to get X-sync-status and X-sync-token and add them to the already captured headers.
		tail -4 logs.zip|sed -n 's/.*X-sync/\nX-sync/p'|grep -a X-sync >> headers.txt
		### Parse captured headers
		shopt -s extglob
		while IFS=':' read key value; do
		    # trim whitespace in "value"
		    value=${value##+([[:space:]])}; value=${value%%+([[:space:]])}
		    case "$key" in
		        Server) httpServer="$value"
		                ;;
		        Content-Type) CT="$value"
		                ;;
		        HTTP*) read httpProto httpStatus httpMsg <<< "$key{$value:+:$value}"
		                ;;
		        Content-Disposition) fileName=$(echo $value | sed -e 's/^.*filename=\"//g;s/\".*$//g')
		        		;;
		        Retry-After) retryAfter="$value"
		        		;;
		        X-sync-status) syncStatus="$value"
		        		;;
		        X-sync-token) syncToken="$value"
		        		;;
		     esac
		done < <(cat headers.txt)
		### Display relevant header information on console
		echo "Resp Code: $httpStatus"
		echo "Retry after: $retryAfter"
		echo "Sync status: $syncStatus"
		echo "Sync token: $syncToken"
		### If response code was 200 make decisions based on sync status and sync token
		if [[ "$httpStatus" == "200" ]]; then
			if [[ "$syncStatus" == "more" && "$lastToken" != "$syncToken" ]]; then
			### If sync status is "more" then continue to download with the provided sync token
				echo "200 OK. Data retrieved. More data immediately available. Will send an additional request to fetch more data."
				mv logs.zip ${fileName}
				echo "${syncToken}" > lasttoken
				logsDownloaded="yes"
			elif [[ "$syncStatus" == "done" && "$lastToken" != "$syncToken" ]]; then
			### If sync status is "done" and a new sync token was provided, use new token in next request at a later time.
				echo "200 OK. Data retrieved. No more data currently available."
				mv logs.zip ${fileName}
				echo "${syncToken}" > lasttoken
				logsDownloaded="yes"
			elif [[ "$syncStatus" == "done" && "$lastToken" == "$syncToken" ]]; then
			### If sync status is "done" and the same sync token was returned, then there was no new data available and the same token should be used in the next request.
				echo "200 OK. Request successful. No data currently available."
				rm -f logs.zip
				echo "${syncToken}" > lasttoken
			elif [[ "$syncStatus" == "abort" && "$lastToken" == "$syncToken" ]]; then
			### If sync status is "abort", then the service has an issue and the request can be sent again using the same sync token.
				echo "200 OK. Data is available, but the service experienced an internal issue. Will try again."
				rm -f logs.zip
				syncStatus="more"
				echo "${syncToken}" > lasttoken
			fi
		elif [[ "$httpStatus" == "400" ]]; then
		### Status 400. Bad request. Send next request without a sync token.
			[ ! -z "$debugMode" ] && echo "400 Bad Request. The start date is later than the end date; bad syntax; or some other request error."			
			syncStatus="none"
			rm -f logs.zip
			[ -e "${destDownloadDir}/lasttoken" ] && rm -f lasttoken
		elif [[ "$httpStatus" == "410" ]]; then
		### The provided token was invalid or expired. Send next request without a sync token.
			[ ! -z "$debugMode" ] && echo "410 Gone. The token is no longer within the given dates or the token references expired data and the token’s associated Web Security Service data is no longer available. The token is no longer valid and is discarded. The client might obtain a new token by requesting a new start date without a token."			
			syncStatus="none"
			rm -f logs.zip
			[ -e "${destDownloadDir}/lasttoken" ] && rm -f lasttoken
		elif [[ "$httpStatus" == "429" ]]; then
		### Request was too early and needs to wait the amount of seconds sent in the Retry-After header. Expected download interval is 5 minutes.
			[ ! -z "$debugMode" ] && echo "429 Too Many Requests. The Web Security Service is unwilling to service the client until a reasonable time has elapsed. Retry after ${retryAfter} seconds."			
			syncStatus="none"
			rm -f logs.zip
		elif [[ "$httpStatus" == "500" ]]; then
		### Service internal error. Send the same request at a later time.
			[ ! -z "$debugMode" ] && echo "500 Internal Error. An internal Web Security Service error prevented the download. The client might need to wait a while before repeating the request."			
			syncStatus="none"
			rm -f logs.zip
		elif [[ "$httpStatus" == "503" ]]; then
		### SyncAPI currently unavailable. Retry later.
			[ ! -z "$debugMode" ] && echo "503 Service Unavailable. No resources are currently available to service the request; or a Web Security Service component is tem- porarily offline. The service might restrict the total number of downloads currently in progress. Wait before repeating the request."			
			syncStatus="none"
			rm -f logs.zip
		fi
	done
	rm -f headers.txt
else
	[ ! -z "$debugMode" ] && echo "No cloud API credentials provided. Process local logs only."
fi

## Unzip log files
for log_archive in $logArchives
do
	if [[ "${log_archive}" == "${logArchives}" ]]; then
		[ ! -z "$debugMode" ] && echo "### No log archives found. Nothing to unzip."
	else
		[ ! -z "$debugMode" ] && echo "### Unzip log file archive: ${log_archive}"
		unzip ${log_archive}
		[ ! -z "$debugMode" ] && echo "Done."; echo "### Delete ${log_archive}"
		rm -f ${log_archive}
		newestFile="$(ls -t *.log.gz| head -1)"
		currentCloudFormat="$(zcat ${newestFile}|head -n 10|grep '#Fields:')"
		echo "$currentCloudFormat" > cloud-log-format.txt
	fi
done
## Process downloaded log files
cd ${scriptDir}
if [[ ${downloadOnly} != "yes" ]]; then
	### Process downloaded log files and output to defined destination file. If sourceID (prefix) provided, add to beginning of each log entry, otherwise output as is. 
	[ ! -z "$debugMode" ] && echo "### Process log files in ${destDownloadDir}/"
	for f in $logFiles
	do
		if [[ "${f}" == "${logFiles}" ]]; then
			[ ! -z "$debugMode" ] && echo "### No new log files found. Nothing to process."
		else
			[ ! -z "$debugMode" ] && echo "Parsing file: $f"
			fileName="$f"
			[ ! -z "$sourceID" ] && zcat $fileName|grep -v "^#" | sed -e 's/^/'"$sourceID"' /' >> $outputFile || zcat $fileName|grep -v "^#" >> $outputFile
		fi
	done
	
	### Archive original log files
	[ ! -z "$debugMode" ] && echo "### Archive log files in ${destArchiveDir}/"
	for fa in $logFiles
	do
		if [[ "${fa}" == "${logFiles}" ]]; then
			[ ! -z "$debugMode" ] && echo "### No new log files found. Nothing to archive."
		else
			[ ! -z "$debugMode" ] && echo "Archiving file: ${fa}"
			mv -f "${fa}" "${destArchiveDir}/"
		fi
	done
	### Remove old log files if archive time set to a given amount of days.
	if [[ ${archiveDays} != "" ]]; then
		[ ! -z "$debugMode" ] && echo "### Removing files in ${destArchiveDir}/ older than ${archiveDays} days"
		find "${destArchiveDir}" -name "*.log.gz" -type f -mtime +"${archiveDays}" -delete
	fi
fi
## Finish
echo "Done."

